
#ifndef PYRED_DEFS_H
#define PYRED_DEFS_H

#include <cstdint>
#include <iostream>
#include <string>
#include <utility>
#include <vector>


using std::uint32_t;
using std::uint64_t;


// hash functions

namespace std {
  template<typename T>
  struct hash<vector<T>>
  {
    size_t operator()(const vector<T>& vec) const {
      size_t h{0};
      const auto elemhasher = hash<T>{};
      for (const auto& elem: vec) {
        h ^= elemhasher(elem) + 0x9e3779b9 + (h << 6) + (h >> 2);
      }
      return h;
    }
  };
  template<typename T1, typename T2>
  struct hash<pair<T1,T2>>
  {
    size_t operator()(const pair<T1,T2>& p) const {
      size_t h{0};
      h ^= hash<T1>{}(p.first) + 0x9e3779b9 + (h << 6) + (h >> 2);
      h ^= hash<T2>{}(p.second) + 0x9e3779b9 + (h << 6) + (h >> 2);
      return h;
    }
  };
}

namespace pyred {

class init_error: public std::exception {
private:
  std::string msg;
public:
  inline init_error(const std::string& s): msg(s) {}
  virtual inline const char* what() const noexcept {
    return msg.c_str();
  }
};

class parser_error: public std::exception {
public:
  std::string msg;
  inline parser_error(const std::string& s): msg(s) {}
  virtual inline const char* what() const noexcept {
    return msg.c_str();
  }
};

class zero_modular_division: public std::exception {
public:
  inline zero_modular_division() {}
  virtual inline const char* what() const noexcept {
    return "Modular arithmetic lead to a spurious zero division.";
  }
};

class input_error: public std::exception {
private:
  std::string msg;
public:
  inline input_error(const std::string& s): msg(s) {}
  virtual inline const char* what() const noexcept {
    return msg.c_str();
  }
};

// stream insertion operators

template<typename T1, typename T2>
std::ostream& operator<<(std::ostream&, const std::pair<T1,T2>&);

template<typename T>
inline std::ostream& operator<<(std::ostream& out, const std::vector<T>& vec) {
  out << "[";
  for (auto it = vec.cbegin(); it != vec.cend(); ++it) {
    if (it != vec.cbegin()) out << ",";
    out << *it;
  }
  out << "]";
  return out;
}

template<typename T1, typename T2>
inline std::ostream& operator<<(std::ostream& out, const std::pair<T1,T2>& p) {
  out << "<" << p.first << "," << p.second << ">";
  return out;
}

// type aliases for the integral class component

class Integral;

using pow_type = int;
using weight_type = uint64_t;
using vs_equation = std::vector<std::pair<Integral,std::string>>;
using ws_equation = std::vector<std::pair<weight_type,std::string>>;
template<typename Coeff>
using wi_equation = std::vector<std::pair<weight_type,Coeff>>;

// type aliases for the solver component

// using intid = uint32_t;
using intid = weight_type;
template<typename Coeff>
using icpair = std::pair<intid,Coeff>;
template<typename Coeff>
using icpairs = std::vector<icpair<Coeff>>;
using eqdata = std::vector<std::pair<intid,std::string>>;

}

#endif
